#
# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Josef Skladanka <jskladan@redhat.com>

import os
import shutil
import subprocess
import re
import rpmUtils.miscutils
import autoqa.koji_utils
import autoqa.util
import koji
from autotest_lib.client.bin import utils
from autoqa.repoinfo import repoinfo
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autoqa.config import getbool

class initscripts(AutoQATest):
    version = 1 # increment if setup() changes

    @ExceptionCatcher()
    def setup(self, **kwargs):
        """
        adds beakerlib repository and installs beakerlib
        """
        retval = subprocess.call(["grep", "-R", "http://afri.fedorapeople.org/beakerlib/", "/etc/yum.repos.d/"])
        if retval != 0:
            f = open("/etc/yum.repos.d/beakerlib.repo", "w")
            f.write("[beakerlib-testing]\nname=Testing repo for BeakerLib\nbaseurl=http://afri.fedorapeople.org/beakerlib/\nenabled=1\ngpgcheck=0\n")
            f.close()
        utils.system('yum -y install beakerlib')

    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.rpmdir = os.path.join(self.tmpdir, 'rpms')
        os.makedirs(self.rpmdir)

    def run_beakerlib_test(self, name, path_to_runtest):
        #install required packages
        """
        there are lines like
          @echo "Requires:        openssh" >> $(METADATA)
        in the Makefile, specifying which packages are needed to run the respective test.
        this code extracts package names from these lines and installs the packages

        We also want to skip the "Requires" line with the package name, since it was installed
         from koji. Makefile has "RunFor" line, which specifies this package name.

        Returns: (result, outputs)
        """
        makefile = os.path.join(path_to_runtest, "Makefile")
        if (os.path.isfile(makefile)):
            #grab the Requires lines
            req_lines = filter(lambda x: x.find("Requires:") > -1, open(makefile, 'r').readlines())
            requires = map(lambda x: re.findall(".*Requires:\s*([^\"']+).*", x)[0], req_lines)

            #grab the RunFor lines
            run_for_lines = filter(lambda x: x.find("RunFor:") > -1, open(makefile, 'r').readlines())
            run_for = map(lambda x: re.findall(".*RunFor:\s*([^\"']+).*", x)[0], run_for_lines)

            #filter out RunFor from Requires, as RunFor was installed from koji
            requires = filter(lambda x: x not in run_for, requires)
            if len(requires) > 0:
                utils.system("yum -y install %s" % " ".join(requires))

        #run the test
        cmd = os.path.join(path_to_runtest, "runtest.sh")
        cmd_result = utils.run(cmd, ignore_status = True, stdout_tee = utils.TEE_TO_LOGS)
        result = 'PASSED' if cmd_result.exit_status == 0 else 'FAILED'
        outputs = cmd_result.stdout

        #store the log journal created by beakerlib
        """
        filters lines containing information about journal files (logs from the test script)
        :: [15:22:06] ::  JOURNAL XML: /tmp/beakerlib-CoEg7JM/journal.xml
        :: [15:22:06] ::  JOURNAL TXT: /tmp/beakerlib-CoEg7JM/journal.txt
        extracts paths to these files, ands stores the files in self.resultsdir
        """
        journal_lines = filter(lambda x: x.find("JOURNAL") > -1, outputs.splitlines())
        journal_files = map(lambda x: x.split(':')[-1].strip(), journal_lines)
        for fname in journal_files:
            shutil.copy(fname, os.path.join(self.resultsdir, "journal_%s.%s" % (name, fname.split('.')[-1])))

        return (result, outputs)

    @ExceptionCatcher()
    def run_once(self, kojitag, **kwargs):
        super(self.__class__, self).run_once()
        if kwargs['hook'] == 'post-koji-build':
            nvrs = [kwargs['nvr']]
            update_id = kwargs['nvr']
        elif kwargs['hook'] == 'post-bodhi-update':
            nvrs = kwargs['nvrs']
            update_id = kwargs['name'] or kwargs['id']

        self.result = 'PASSED'
        self.results_count = {'PASSED': 0, 'FAILED': 0} # count how many of which results do we have
        self.tests_passed = []
        self.tests_failed = []
        self.outputs = []
        self.highlights = []

        for nvr in nvrs:
            # add header
            msg = '%s\n%s\n%s' % ('='*40, nvr, '='*40)
            print msg
            self.outputs.append(msg)
            #find all tests for package $name
            nvrea = rpmUtils.miscutils.splitFilename(nvr + '.noarch')
            name = nvrea[0]
            testdir = os.path.join(self.bindir, "./tests/%s" % name)
            assert os.path.isdir(testdir), "No initscript checker found for package %s" % name
            tests = []
            """
            walks through ./tests/$NAME and all it's subdirs. Every time it finds
            file runtest.sh, it stores a tuple(test_name, path) to tests[],
            where test_name is 'openssh' for ./tests/openssh, openssh_subdir for ./tests/openssh/subdir, etc.
            """
            for (path, dirs, files) in os.walk(testdir):
                if "runtest.sh" in files:
                    test_name = path.replace(os.path.join(self.bindir, "./tests/"), "").replace(os.path.sep, "_")
                    tests.append((test_name, path))

            #install packages from koji
            koji = autoqa.koji_utils.SimpleKojiClientSession()
            pkgurls = koji.nvr_to_urls(nvr, arches = os.uname()[-1])
            rpms = []
            print "Saving RPMs to %s" % self.rpmdir
            #download packages
            for p in pkgurls:
                # fetch package to rpmdir
                print "Grabbing %s" % p
                localfile = os.path.join(self.rpmdir, os.path.basename(p))
                autoqa.util.grabber.urlgrab(p, localfile)
                rpms.append(localfile)
            #and install them
            cmd = "yum -y --nogpgcheck localinstall %s" % " ".join(rpms)
            utils.system_output(cmd)

            #sort tests alphabetically by test_name
            tests.sort(key = lambda x: x[0])

            # run tests
            for test in tests:
                msg = "RUNNING: %s" % test[0]
                msg += '\n%s' % ("="*len(msg),)
                print msg
                self.outputs.append(msg)

                (result, outputs) = self.run_beakerlib_test(test[0], test[1])
                self.results_count[result] += 1
                self.outputs.append(outputs)

                if result == 'PASSED':
                    self.tests_passed.append(test[0])
                else:
                    self.tests_failed.append(test[0])

            # email results to mailing list and to pkg owner if they optin
            repo = repoinfo.getrepo_by_tag(kojitag)
            send_optin_email = getbool(self.config.get('notifications', 'send_optin_email'))
            if (repo is not None and send_optin_email and
                autoqa.util.check_opt_in(name, repo['collection_name'])):
                #FIXME - hardcoded partial email address here - obviously sub-par
                self.mail_to.append('%s-owner@fedoraproject.org' % name)


        # set result variables
        if len(self.tests_failed) > 0:
            self.result = 'FAILED'

        self.highlights.append("Following tests passed: %s" % ', '.join(self.tests_passed))
        self.highlights.append("Following tests failed: %s" % ', '.join(self.tests_failed))

        self.summary = "%d PASSED, %d FAILED for %s" % (self.results_count['PASSED'], self.results_count['FAILED'], update_id)

