# -*- coding: utf-8 -*-
#
# Copyright 2009, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Kamil Paral <kparal@redhat.com>

from autotest_lib.client.bin import utils
import autoqa.koji_utils
import autoqa.util
from autoqa.repoinfo import repoinfo
import os
import rpmUtils.miscutils
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autoqa.config import getbool

class rpmguard(AutoQATest):
    version = 1 # increment if setup() changes

    @ExceptionCatcher()
    def setup(self, **kwargs):
        utils.system('yum -y install rpmlint')

    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.rpmguard = os.path.join(self.bindir, 'rpmguard')
        self.rpmdir = os.path.join(self.tmpdir, 'rpms')
        os.makedirs(self.rpmdir)

    @ExceptionCatcher()
    def run_once(self, kojitag, **kwargs):
        super(self.__class__, self).run_once()
        if kwargs['hook'] == 'post-koji-build':
            nvrs = [kwargs['nvr']]
            update_id = kwargs['nvr']
        elif kwargs['hook'] == 'post-bodhi-update':
            nvrs = kwargs['nvrs']
            update_id = kwargs['name'] or kwargs['id']

        self.result = 'PASSED'
        # order for evaluation of final result; higher index means preference
        self.result_order = ('PASSED','INFO','FAILED','ABORTED')
        self.nvr_results = {} # results for invidual packages
        self.outputs = []
        self.highlights = []

        for nvr in nvrs:
            # add header
            msg = '%s\n%s\n%s' % ('='*40, nvr, '='*40)
            print msg
            self.outputs.append(msg)
            # run the test for this nvr
            (result, highlights, outputs, warn_count) = self.test_nvr(nvr, kojitag)
            # collect output
            self.nvr_results[nvr] = result
            if self.result_order.index(result) > self.result_order.index(self.result):
                self.result = result
            if highlights:
                self.highlights.append(highlights)
            self.outputs.append(outputs)
            # add footer
            msg = 'N: ----'
            print msg
            if (self.outputs and self.outputs[-1].splitlines() and
                self.outputs[-1].splitlines()[-1] != msg):
                self.outputs.append(msg)

            # email results to mailing list and to pkg owner if they optin
            repo = repoinfo.getrepo_by_tag(kojitag)
            pkg_name = rpmUtils.miscutils.splitFilename(nvr + '.noarch')[0]
            send_optin_email = getbool(self.config.get('notifications', 'send_optin_email'))
            if repo is not None and send_optin_email and \
            autoqa.util.check_opt_in(pkg_name, repo['collection_name']):
                #FIXME - hardcoded partial email address here - obviously sub-par
                self.mail_to.append('%s-owner@fedoraproject.org' % pkg_name)

        # create self.summary
        if kwargs['hook'] == 'post-koji-build':
            if result != 'ABORTED':
                self.summary = '%d warnings for %s' % (warn_count, update_id)
            else:
                self.summary = update_id
        elif kwargs['hook'] == 'post-bodhi-update':
            # create result line like "1 PASSED, 2 FAILED, 3 INFO"
            result_count = []
            for res in self.result_order:
                if res in self.nvr_results.values():
                    count = len([k for k in self.nvr_results.keys() if self.nvr_results[k] == res])
                    result_count.append('%d %s' % (count, res))
            result_count = ', '.join(result_count)
            self.summary = '%s for %s' % (result_count, update_id)


    def test_nvr(self, nvr, kojitag):
        '''
        Test a single NVR.
        Returns (result, highlights, outputs, warn_count).
        '''
        result = 'PASSED'
        highlights = []
        outputs = []
        warn_count = 0

        def get_result():
            '''Format the variables as to be returned'''
            warn_count = len([line for line in outputs if line.startswith('W: ')])
            return (result, '\n'.join(highlights), '\n'.join(outputs), warn_count)

        # temporary results for each architecture
        arch_results = {}
        koji = autoqa.koji_utils.SimpleKojiClientSession()

        # get the most recent release available
        # add .noarch to parse filename correctly
        nvrea = rpmUtils.miscutils.splitFilename(nvr + '.noarch')
        name = nvrea[0]
        lastBuild = koji.list_previous_release(name, kojitag,
            max_evr=(nvrea[3], nvrea[1], nvrea[2]))

        # if there is no such build, we don't have anything to compare
        if not lastBuild:
            msg = "N: There is no previous build of %s in %s tag (or its parents)." % (nvr, kojitag)
            print msg
            outputs.append(msg)
            return get_result()

        # now we need list of RPMs available for each build
        new_rpms = koji.nvr_to_rpms(nvr, src=False)
        old_rpms = koji.nvr_to_rpms(lastBuild['nvr'], src=False)
        # and match the RPMs according to build name and architecture as
        # (old one, new one)
        rpm_to_match = []

        # traverse all RPMs in the newer build
        for new_rpm in new_rpms:
            # find the matching older RPM, there should be exactly one
            old_rpm = [r for r in old_rpms if r['arch'] == new_rpm['arch']
                        and r['name'] == new_rpm['name']]
            # we should mark all these packages as already matched
            for o in old_rpm:
                assert not o.has_key('seen'), 'The foreign object already has the same attribute as we use'
                o['seen'] = True
            # older RPM may not exist
            if not old_rpm:
                msg = 'N: There is no previous RPM for %s.%s' % (new_rpm['nvr'], new_rpm['arch'])
                print msg
                outputs.append(msg)
                continue
            # there certainly shouldn't be more than 1 matching older RPM, if there is,
            # we have a bug somewhere
            assert len(old_rpm) <= 1, 'There are %d existing older RPMs for %s.%s: %s' % \
                (len(old_rpm), new_rpm['nvr'], new_rpm['arch'], old_rpm)

            # now there is certainly one older RPM matching the newer one
            old_rpm = old_rpm[0]
            rpm_to_match.append((old_rpm, new_rpm))

        # there may be some older RPMs not longer available in the newer build
        for old_rpm in old_rpms:
            if not old_rpm.has_key('seen'):
                msg = 'N: There is no new RPM to match %s.%s' % (old_rpm['nvr'], old_rpm['arch'])
                print msg
                outputs.append(msg)

        # let's compare the packages finally
        for old_rpm, new_rpm in rpm_to_match:
            # fetch old and new rpms to rpmdir
            url = old_rpm['url']
            print "Grabbing %s" % url
            old_file = os.path.join(self.rpmdir, os.path.basename(url))
            autoqa.util.grabber.urlgrab(url, old_file)
            url = new_rpm['url']
            print "Grabbing %s" % url
            new_file = os.path.join(self.rpmdir, os.path.basename(url))
            autoqa.util.grabber.urlgrab(url, new_file)

            # run rpmguard
            cmd = '%s %s %s' % (self.rpmguard, old_file, new_file)
            cmd_result = utils.run(cmd+" 2>&1", ignore_status=True,
                stdout_tee=utils.TEE_TO_LOGS, stderr_tee=utils.TEE_TO_LOGS)
            # store the results in temporary arch_results
            key = (old_rpm['nvr'],new_rpm['nvr'])
            if not arch_results.has_key(key):
                arch_results[key] = {}
            archs = arch_results[key]
            archs[old_rpm['arch']] = cmd_result.stdout
            # check command return code
            if cmd_result.exit_status == 0:
                result = 'PASSED'
            elif cmd_result.exit_status == 10:
                result = 'INFO'
            else:
                result = 'ABORTED'

        # Copy arch_results to results and eliminate duplicates
        for (old_nvr, new_nvr) in arch_results.keys():
            res_arch = {}
            for arch, res in arch_results[(old_nvr, new_nvr)].iteritems():
                if not res_arch.has_key(res):
                    res_arch[res] = [arch]
                else:
                    res_arch[res].append(arch)
            for res, archs in res_arch.iteritems():
                archlist = ''
                for arch in archs:
                    archlist += arch + ', '
                archlist = archlist[:-2]
                outputs.append('N: Comparing %s and %s (archs: %s) ...' % \
                    (old_nvr, new_nvr, archlist))
                if res:
                    outputs.append(res.rstrip()) #strip newline at the end
                outputs.append('N: ----')

        return get_result()
